# Copyright (C) 2018-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.db import models

from swh.web.utils.typing import SaveOriginRequestInfo


class SaveAuthorizedOrigin(models.Model):
    """
    Model table holding origin urls authorized to be loaded into the archive.
    """

    url = models.CharField(max_length=200, null=False)

    class Meta:
        app_label = "swh_web_save_code_now"
        db_table = "save_authorized_origin"
        indexes = [models.Index(fields=["url"])]

    def __str__(self):
        return self.url


class SaveUnauthorizedOrigin(models.Model):
    """
    Model table holding origin urls not authorized to be loaded into the
    archive.
    """

    url = models.CharField(max_length=200, null=False)

    class Meta:
        app_label = "swh_web_save_code_now"
        db_table = "save_unauthorized_origin"
        indexes = [models.Index(fields=["url"])]

    def __str__(self):
        return self.url


SAVE_REQUEST_ACCEPTED = "accepted"
SAVE_REQUEST_REJECTED = "rejected"
SAVE_REQUEST_PENDING = "pending"

SAVE_REQUEST_STATUS = [
    (SAVE_REQUEST_ACCEPTED, SAVE_REQUEST_ACCEPTED),
    (SAVE_REQUEST_REJECTED, SAVE_REQUEST_REJECTED),
    (SAVE_REQUEST_PENDING, SAVE_REQUEST_PENDING),
]

SAVE_TASK_NOT_CREATED = "not created"
SAVE_TASK_NOT_YET_SCHEDULED = "not yet scheduled"
SAVE_TASK_SCHEDULED = "scheduled"
SAVE_TASK_SUCCEEDED = "succeeded"
SAVE_TASK_FAILED = "failed"
SAVE_TASK_RUNNING = "running"

SAVE_TASK_STATUS = [
    (SAVE_TASK_NOT_CREATED, SAVE_TASK_NOT_CREATED),
    (SAVE_TASK_NOT_YET_SCHEDULED, SAVE_TASK_NOT_YET_SCHEDULED),
    (SAVE_TASK_SCHEDULED, SAVE_TASK_SCHEDULED),
    (SAVE_TASK_SUCCEEDED, SAVE_TASK_SUCCEEDED),
    (SAVE_TASK_FAILED, SAVE_TASK_FAILED),
    (SAVE_TASK_RUNNING, SAVE_TASK_RUNNING),
]

VISIT_STATUS_CREATED = "created"
VISIT_STATUS_ONGOING = "ongoing"
VISIT_STATUS_FULL = "full"
VISIT_STATUS_PARTIAL = "partial"
VISIT_STATUS_NOT_FOUND = "not_found"
VISIT_STATUS_FAILED = "failed"

VISIT_STATUSES = [
    (VISIT_STATUS_CREATED, VISIT_STATUS_CREATED),
    (VISIT_STATUS_ONGOING, VISIT_STATUS_ONGOING),
    (VISIT_STATUS_FULL, VISIT_STATUS_FULL),
    (VISIT_STATUS_PARTIAL, VISIT_STATUS_PARTIAL),
    (VISIT_STATUS_NOT_FOUND, VISIT_STATUS_NOT_FOUND),
    (VISIT_STATUS_FAILED, VISIT_STATUS_FAILED),
]


class SaveOriginRequest(models.Model):
    """
    Model table holding all the save origin requests issued by users.
    """

    id = models.BigAutoField(primary_key=True)
    request_date = models.DateTimeField(auto_now_add=True)
    visit_type = models.CharField(max_length=200, null=False)
    visit_status = models.TextField(choices=VISIT_STATUSES, null=True)
    origin_url = models.CharField(max_length=200, null=False)
    status = models.TextField(choices=SAVE_REQUEST_STATUS, default=SAVE_REQUEST_PENDING)
    loading_task_id = models.IntegerField(default=-1)
    visit_date = models.DateTimeField(null=True)
    loading_task_status = models.TextField(
        choices=SAVE_TASK_STATUS, default=SAVE_TASK_NOT_CREATED
    )
    # store ids of users that submitted the request as string list
    user_ids = models.TextField(null=True)
    note = models.TextField(null=True)
    from_webhook = models.BooleanField(default=False)
    webhook_origin = models.CharField(max_length=200, null=True)
    # if None, no try to retrieve the snapshot has been performed
    # if empty string, try to retrieve the snapshot has been performed but none was found
    snapshot_swhid = models.CharField(max_length=200, null=True)

    class Meta:
        app_label = "swh_web_save_code_now"
        db_table = "save_origin_request"
        ordering = ["-id"]
        indexes = [models.Index(fields=["origin_url", "status"])]

    def to_dict(self) -> SaveOriginRequestInfo:
        """Map the request save model object to a json serializable dict.

        Returns:
            The corresponding SaveOriginRequetsInfo json serializable dict.

        """
        visit_date = self.visit_date
        return SaveOriginRequestInfo(
            id=self.id,
            origin_url=self.origin_url,
            visit_type=self.visit_type,
            save_request_date=self.request_date.isoformat(),
            save_request_status=self.status,
            save_task_status=self.loading_task_status,
            visit_status=self.visit_status,
            visit_date=visit_date.isoformat() if visit_date else None,
            loading_task_id=self.loading_task_id,
            note=self.note,
            from_webhook=self.from_webhook,
            webhook_origin=self.webhook_origin,
            snapshot_swhid=self.snapshot_swhid or None,
        )

    def __str__(self) -> str:
        return str(self.to_dict())



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import contextmanager
from dataclasses import dataclass
from email.message import EmailMessage
from io import BytesIO, StringIO
import re
import sys
from typing import Callable, Iterator
from unittest.mock import MagicMock

import pytest

from django.core.management import call_command
from django.dispatch import Signal

from swh.web.inbound_email.signals import EmailProcessingStatus, email_received


class MockedStdin:
    def __init__(self):
        self.buffer = BytesIO()


@dataclass
class CommandReturn:
    out: str
    err: str


@contextmanager
def signal_receiver(signal: Signal, name: str = "receiver_name") -> Iterator[Callable]:
    receiver = MagicMock()
    receiver.configure_mock(__name__=name, __qualname__=name)

    try:
        signal.connect(receiver)
        yield receiver
    finally:
        signal.disconnect(receiver)


def call_process_inbound_email(stdin_data: bytes) -> CommandReturn:
    orig_stdin = sys.stdin
    try:
        sys.stdin = MockedStdin()  # type: ignore
        sys.stdin.buffer.write(stdin_data)
        sys.stdin.buffer.seek(0)

        out = StringIO()
        err = StringIO()

        call_command("process_inbound_email", stdout=out, stderr=err)

        out.seek(0)
        err.seek(0)
        return CommandReturn(out=out.read(), err=err.read())
    finally:
        sys.stdin = orig_stdin


def test_empty_stdin(caplog):
    ret = call_process_inbound_email(b"")
    assert ret.out == ""
    assert ret.err == ""

    assert len(caplog.records) == 1
    log = caplog.records[0]
    assert log.levelname == "ERROR"
    assert "Unhandled message" in log.getMessage()


@pytest.mark.parametrize(
    "return_value,err_contents",
    [
        # When the email gets processed by one of the receivers, the management command
        # should not emit any output.
        (EmailProcessingStatus.PROCESSED, ""),
        # When a receiver fails, the management command outputs a message to this effect
        (EmailProcessingStatus.FAILED, "Failed receiver.*receiver_name"),
        # When all receivers ignore a message, this fact is printed too
        (EmailProcessingStatus.IGNORED, "Unhandled message"),
    ],
)
def test_signal_receiver(return_value, err_contents, caplog):
    """Check that signal receivers are properly called when running the management command.

    Check for output depending on its return value"""
    message = EmailMessage()
    message["to"] = "test@example.com"
    message["subject"] = "Test Subject"
    message.set_content("This is a test message.\n")

    with signal_receiver(email_received) as receiver:
        receiver.return_value = return_value

        ret = call_process_inbound_email(bytes(message))
        assert ret.out == ""
        assert ret.err == ""
        output = "\n".join(record.getMessage() for record in caplog.records)
        if err_contents:
            assert re.match(err_contents, output)
        else:
            assert output == ""

        calls = receiver.call_args_list

        assert len(calls) == 1
        assert bytes(calls[0][1]["message"]) == bytes(message)


def test_multiple_receivers(caplog):
    message = EmailMessage()
    message["to"] = "test@example.com"
    message["subject"] = "Test Subject"
    message.set_content("This is a test message.\n")

    with signal_receiver(email_received, name="ignored") as ignored, signal_receiver(
        email_received, name="processed"
    ) as processed:
        ignored.return_value = EmailProcessingStatus.IGNORED
        processed.return_value = EmailProcessingStatus.PROCESSED

        ret = call_process_inbound_email(bytes(message))
        assert ret.out == ""
        assert ret.err == ""

        assert not caplog.records

        for receiver in [ignored, processed]:
            calls = receiver.call_args_list

            assert len(calls) == 1
            assert bytes(calls[0][1]["message"]) == bytes(message)


def test_signal_receiver_exception(caplog):
    message = EmailMessage()
    message["to"] = "test@example.com"
    message["subject"] = "Test Subject"
    message.set_content("This is a test message.\n")

    with signal_receiver(email_received, name="exception_raised") as receiver:
        receiver.side_effect = ValueError("I'm broken!")

        ret = call_process_inbound_email(bytes(message))
        assert ret.out == ""
        assert ret.err == ""

        output = "\n".join(
            record.getMessage() + ("\n" + record.exc_text if record.exc_text else "")
            for record in caplog.records
        )

        assert re.match("Failed receiver.*exception_raised", output)
        assert "following exception" in output
        assert "ValueError" in output
        assert "I'm broken" in output



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Any, Dict, Optional

from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect, render

from swh.model.swhids import ObjectType
from swh.web.browse.browseurls import browse_route
from swh.web.browse.snapshot_context import get_snapshot_context
from swh.web.browse.utils import gen_link, get_directory_entries, get_readme_to_display
from swh.web.utils import archive, gen_path_info, reverse, swh_object_icons
from swh.web.utils.exc import (
    NotFoundExc,
    http_status_code_message,
    sentry_capture_exception,
)
from swh.web.utils.identifiers import get_swhids_info
from swh.web.utils.typing import DirectoryMetadata, SWHObjectInfo


def _directory_browse(
    request: HttpRequest, sha1_git: str, path: Optional[str] = None
) -> HttpResponse:
    root_sha1_git = sha1_git
    dir_sha1_git: Optional[str] = sha1_git
    target_type = None
    error_info: Dict[str, Any] = {"status_code": 200, "description": None}
    if path:
        try:
            dir_info = archive.lookup_directory_with_path(sha1_git, path)
            dir_sha1_git = dir_info["target"]
            target_type = dir_info["type"]
        except NotFoundExc as e:
            error_info["status_code"] = 404
            error_info["description"] = f"NotFoundExc: {str(e)}"
            dir_sha1_git = None

    if target_type == "file":
        params = request.GET.dict()
        if "origin_url" not in params:
            # no origin context, prepend root directory id to path for
            # breadcrumbs navigation
            params["path"] = f"{sha1_git}/{path}"
        browse_content_url = reverse(
            "browse-content",
            url_args={"query_string": f"sha1_git:{dir_sha1_git}"},
            query_params=params,
        )
        return redirect(browse_content_url)

    dirs, files = [], []
    if dir_sha1_git is not None:
        dirs, files = get_directory_entries(dir_sha1_git)
    origin_url = request.GET.get("origin_url")
    if not origin_url:
        origin_url = request.GET.get("origin")
    snapshot_id = request.GET.get("snapshot")
    snapshot_context = None
    if origin_url is not None or snapshot_id is not None:
        try:
            snapshot_context = get_snapshot_context(
                snapshot_id=snapshot_id,
                origin_url=origin_url,
                branch_name=request.GET.get("branch"),
                release_name=request.GET.get("release"),
                revision_id=request.GET.get("revision"),
                path=path,
            )
        except NotFoundExc as e:
            if str(e).startswith("Origin") and origin_url is not None:
                raw_dir_url = reverse(
                    "browse-directory", url_args={"sha1_git": dir_sha1_git}
                )
                error_message = (
                    "The Software Heritage archive has a directory "
                    "with the hash you provided but the origin "
                    "mentioned in your request appears broken: %s. "
                    "Please check the URL and try again.\n\n"
                    "Nevertheless, you can still browse the directory "
                    "without origin information: %s"
                    % (gen_link(origin_url), gen_link(raw_dir_url))
                )
                raise NotFoundExc(error_message)
            else:
                raise e

    path_info = gen_path_info(path)

    query_params = snapshot_context["query_params"] if snapshot_context else {}

    breadcrumbs = []
    breadcrumbs.append(
        {
            "name": root_sha1_git[:7],
            "url": reverse(
                "browse-directory",
                url_args={"sha1_git": root_sha1_git},
                query_params={**query_params, "path": None},
            ),
        }
    )

    for pi in path_info:
        breadcrumbs.append(
            {
                "name": pi["name"],
                "url": reverse(
                    "browse-directory",
                    url_args={"sha1_git": root_sha1_git},
                    query_params={
                        **query_params,
                        "path": pi["path"],
                    },
                ),
            }
        )

    path = "" if path is None else (path + "/")

    for d in dirs:
        if d["type"] == "rev":
            d["url"] = reverse(
                "browse-revision",
                url_args={"sha1_git": d["target"]},
                query_params=query_params,
            )
        else:
            d["url"] = reverse(
                "browse-directory",
                url_args={"sha1_git": root_sha1_git},
                query_params={
                    **query_params,
                    "path": path + d["name"],
                },
            )

    sum_file_sizes = 0

    readmes = {}

    for f in files:
        query_string = "sha1_git:" + f["target"]
        f["url"] = reverse(
            "browse-content",
            url_args={"query_string": query_string},
            query_params={
                **query_params,
                "path": root_sha1_git + "/" + path + f["name"],
            },
        )
        if f["length"] is not None:
            sum_file_sizes += f["length"]
        if f["name"].lower().startswith("readme"):
            readmes[f["name"]] = f["checksums"]["sha1"]

    readme_name, readme_url, readme_html = get_readme_to_display(readmes)

    dir_metadata = DirectoryMetadata(
        object_type=ObjectType.DIRECTORY,
        object_id=dir_sha1_git,
        directory=root_sha1_git,
        nb_files=len(files),
        nb_dirs=len(dirs),
        sum_file_sizes=sum_file_sizes,
        root_directory=root_sha1_git,
        path=f"/{path}" if path else None,
        revision=None,
        revision_found=None,
        release=None,
        snapshot=None,
    )

    vault_cooking = {
        "directory_context": True,
        "directory_swhid": f"swh:1:dir:{dir_sha1_git}",
        "revision_context": False,
        "revision_swhid": None,
    }

    swh_objects = [
        SWHObjectInfo(object_type=ObjectType.DIRECTORY, object_id=dir_sha1_git)
    ]

    if snapshot_context:
        if snapshot_context["revision_id"]:
            swh_objects.append(
                SWHObjectInfo(
                    object_type=ObjectType.REVISION,
                    object_id=snapshot_context["revision_id"],
                )
            )
        swh_objects.append(
            SWHObjectInfo(
                object_type=ObjectType.SNAPSHOT,
                object_id=snapshot_context["snapshot_id"],
            )
        )
        if snapshot_context["release_id"]:
            swh_objects.append(
                SWHObjectInfo(
                    object_type=ObjectType.RELEASE,
                    object_id=snapshot_context["release_id"],
                )
            )

    swhids_info = get_swhids_info(swh_objects, snapshot_context, dir_metadata)

    heading = "Directory - %s" % dir_sha1_git
    if breadcrumbs:
        dir_path = "/".join([bc["name"] for bc in breadcrumbs]) + "/"
        heading += " - %s" % dir_path

    top_right_link = None
    if (
        snapshot_context is not None
        and not snapshot_context["is_empty"]
        and snapshot_context["revision_id"] is not None
    ):
        history_url = reverse(
            "browse-revision-log",
            url_args={"sha1_git": snapshot_context["revision_id"]},
            query_params=query_params,
        )
        top_right_link = {
            "url": history_url,
            "icon": swh_object_icons["revisions history"],
            "text": "History",
        }

    return render(
        request,
        "browse-directory.html",
        {
            "heading": heading,
            "swh_object_id": swhids_info[0]["swhid"],
            "swh_object_name": "Directory",
            "swh_object_metadata": dir_metadata,
            "dirs": dirs,
            "files": files,
            "breadcrumbs": breadcrumbs,
            "top_right_link": top_right_link,
            "readme_name": readme_name,
            "readme_url": readme_url,
            "readme_html": readme_html,
            "snapshot_context": snapshot_context,
            "vault_cooking": vault_cooking,
            "show_actions": True,
            "swhids_info": swhids_info,
            "error_code": error_info["status_code"],
            "error_message": http_status_code_message.get(error_info["status_code"]),
            "error_description": error_info["description"],
        },
        status=error_info["status_code"],
    )


@browse_route(
    r"directory/(?P<sha1_git>[0-9a-f]+)/",
    view_name="browse-directory",
    checksum_args=["sha1_git"],
)
def directory_browse(request: HttpRequest, sha1_git: str) -> HttpResponse:
    """Django view for browsing the content of a directory identified
    by its sha1_git value.

    The url that points to it is
    :http:get:`/browse/directory/(sha1_git)/`
    """
    return _directory_browse(request, sha1_git, request.GET.get("path"))


@browse_route(
    r"directory/(?P<sha1_git>[0-9a-f]+)/(?P<path>.+)/",
    view_name="browse-directory-legacy",
    checksum_args=["sha1_git"],
)
def directory_browse_legacy(
    request: HttpRequest, sha1_git: str, path: str
) -> HttpResponse:
    """Django view for browsing the content of a directory identified
    by its sha1_git value.

    The url that points to it is
    :http:get:`/browse/directory/(sha1_git)/(path)/`
    """
    return _directory_browse(request, sha1_git, path)


@browse_route(
    r"directory/resolve/content-path/(?P<sha1_git>[0-9a-f]+)/",
    view_name="browse-directory-resolve-content-path",
    checksum_args=["sha1_git"],
)
def _directory_resolve_content_path(
    request: HttpRequest, sha1_git: str
) -> HttpResponse:
    """
    Internal endpoint redirecting to data url for a specific file path
    relative to a root directory.
    """
    try:
        path = os.path.normpath(request.GET.get("path", ""))
        if not path.startswith("../"):
            dir_info = archive.lookup_directory_with_path(sha1_git, path)
            if dir_info["type"] == "file":
                sha1 = dir_info["checksums"]["sha1"]
                data_url = reverse(
                    "browse-content-raw", url_args={"query_string": sha1}
                )
                return redirect(data_url)
    except Exception as exc:
        sentry_capture_exception(exc)
    return HttpResponse(status=404)



#!/usr/bin/env python3
# Copyright (C) 2015-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License v3 or later
# See top-level LICENSE file for more information

import os

from setuptools import setup

# package generated static assets as module data files
data_files = []
for folder in ("static/", "assets/"):
    for root, _, files in os.walk(folder):
        root_files = [os.path.join(root, i) for i in files]
        data_files.append((os.path.join("share/swh/web", root), root_files))

# TODO: data_files usage has been deprecated, we should move away from using it...
setup(
    data_files=data_files,
)
